"""Ontology-focused SHACL and RDF quads validation tests."""

from __future__ import annotations

from importlib import resources
import json
import os
from copy import deepcopy
from pathlib import Path

import pytest
from rdflib import Dataset
from rdflib import Graph
from pyshacl import validate as shacl_validate

from cf_ontology import OntologyManager, ingest_rdf_documents
from cf_ontology.datatypes_mapping import (
    normalize_parameter_type,
    parameter_type_to_literal_datatype,
    normalize_port_literal_datatype,
    supported_port_literal_datatypes,
)
from cf_ontology.quad_store import QuadStore
from cf_ontology.rdf_document import load_rdf_document
from cf_ontology.semantics_paths import resolve_semantics_db_path, resolve_semantics_dir


CF_BASE = "https://cogniflow.odea-project.org/cf#"
CFRUN_BASE = f"{CF_BASE}runner/"
_STRUCTURED_FORMAT = "json" + "-ld"
_CONTEXT_FILENAME = "cf_context." + "json" + "ld"


def _repo_root() -> Path:
    return Path(__file__).resolve().parents[3]


def _resolve_steps_document(module_name: str, repo_fallback: Path) -> Path:
    """
    Prefer the checked-out repository manifest for deterministic repo tests.
    Fall back to installed package resources when the repo path is unavailable.
    """
    if repo_fallback.is_file():
        return repo_fallback
    try:
        candidate = resources.files(module_name) / "steps.nq"
        if candidate.is_file():
            with resources.as_file(candidate) as resolved:
                return Path(resolved).resolve()
    except (ModuleNotFoundError, FileNotFoundError, ImportError):
        pass
    return repo_fallback


def _find_step_by_suffix(steps: list[dict], suffix: str) -> dict:
    for step in steps:
        if str(step.get("@id", "")).endswith(suffix):
            return step
    raise AssertionError(f"Step ending with '{suffix}' not found in DTO payload.")


def _find_named(items: list[dict], key: str, expected: str) -> dict:
    for item in items:
        if str(item.get(key, "")) == expected:
            return item
    raise AssertionError(f"Entry with {key}='{expected}' not found.")


def _extend_string_values(target: list[str], value: object) -> None:
    if isinstance(value, list):
        target.extend(str(item) for item in value)
    elif isinstance(value, str):
        target.append(value)


def _has_type(node: dict, expected: str) -> bool:
    node_type = node.get("@type")
    if isinstance(node_type, str):
        return node_type == expected
    if isinstance(node_type, list):
        return expected in node_type
    return False


def _contains_supported_pipeline_example(document: dict) -> bool:
    graph = document.get("@graph")
    nodes = graph if isinstance(graph, list) else [document]
    for node in nodes:
        if not isinstance(node, dict):
            continue
        if _has_type(node, "cf:ProcessingPipeline") or _has_type(
            node, f"{CF_BASE}ProcessingPipeline"
        ):
            return True
    return False


def _contains_legacy_cogniflow_url(value: object) -> bool:
    if isinstance(value, str):
        legacy_host = "cogniflow" + ".org"
        return legacy_host in value.lower()
    if isinstance(value, list):
        return any(_contains_legacy_cogniflow_url(item) for item in value)
    if isinstance(value, dict):
        return any(_contains_legacy_cogniflow_url(item) for item in value.values())
    return False


def _id_has_suffix(value: object, suffix: str) -> bool:
    return str(value).endswith(suffix)


def _canonical_context() -> dict[str, str]:
    return {
        "adf": "http://allotrope.org/datamodel#",
        "cf": CF_BASE,
        "cfrun": CFRUN_BASE,
        "owl": "http://www.w3.org/2002/07/owl#",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
        "schema": "https://schema.org/",
        "sh": "http://www.w3.org/ns/shacl#",
        "skos": "http://www.w3.org/2004/02/skos/core#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
    }


def _document_from_dataset(dataset: Dataset) -> dict:
    serialized = dataset.serialize(
        format=_STRUCTURED_FORMAT,
        context=_canonical_context(),
        auto_compact=True,
    )
    if isinstance(serialized, bytes):
        serialized = serialized.decode("utf-8")
    payload = json.loads(serialized)
    if isinstance(payload, list):
        return {"@context": _canonical_context(), "@graph": payload}
    if isinstance(payload, dict):
        if "@graph" in payload:
            if "@context" not in payload:
                payload["@context"] = _canonical_context()
            return payload
        return {"@context": _canonical_context(), "@graph": [payload]}
    raise ValueError("Unsupported RDF document payload generated from N-Quads dataset.")


def _load_graph_resource(path: Path) -> Graph:
    if path.suffix.lower() == ".nq":
        dataset = Dataset()
        dataset.parse(str(path), format="nquads")
        graph = Graph()
        for named_graph in dataset.graphs():
            for subj, pred, obj in named_graph.triples((None, None, None)):
                graph.add((subj, pred, obj))
        return graph
    return Graph().parse(str(path), format=_STRUCTURED_FORMAT)


def _load_ontology_resource(relative_parts: tuple[str, ...]) -> dict:
    path = _repo_root().joinpath("sandcastle", "cf_ontology", "src", "cf_ontology", "ontology", *relative_parts)
    if path.suffix.lower() == ".nq":
        dataset = Dataset()
        dataset.parse(str(path), format="nquads")
        return _document_from_dataset(dataset)
    return json.loads(path.read_text(encoding="utf-8"))


def _load_json_document(path: Path) -> dict:
    return load_rdf_document(path, label="test RDF document")


def _cf_ontology_source_path(*parts: str) -> Path:
    return _repo_root().joinpath("sandcastle", "cf_ontology", "src", "cf_ontology", *parts)


def _repo_step_manifest_paths() -> list[Path]:
    root = _repo_root() / "sandcastle" / "cf_basic_steps"
    return [
        root / "cf_basic_dev" / "src" / "cf_basic_dev" / "steps.nq",
        root / "cf_basic_io" / "src" / "cf_basic_io" / "steps.nq",
        root / "cf_basic_signal" / "src" / "cf_basic_signal" / "steps.nq",
        root / "cf_basic_sinks" / "src" / "cf_basic_sinks" / "steps.nq",
    ]


def _is_url(text: str) -> bool:
    lowered = text.lower()
    return lowered.startswith("http://") or lowered.startswith("https://") or lowered.startswith(
        "file://"
    )


def _resolve_relative_context_paths(document: dict, base_dir: Path) -> dict:
    """
    Rewrite relative RDF document @context entries to absolute filesystem paths.
    """
    payload = deepcopy(document)

    def _map_context_value(value):
        if isinstance(value, str):
            if value.replace("\\", "/").endswith(_CONTEXT_FILENAME):
                return _canonical_context()
            if _is_url(value):
                return value
            candidate = Path(value)
            if not candidate.is_absolute():
                candidate = (base_dir / candidate).resolve()
            return str(candidate) if candidate.exists() else value
        if isinstance(value, list):
            return [_map_context_value(item) for item in value]
        if isinstance(value, dict):
            return {key: _map_context_value(inner) for key, inner in value.items()}
        return value

    def _walk(node):
        if isinstance(node, dict):
            out = {}
            for key, value in node.items():
                out[key] = _map_context_value(value) if key == "@context" else _walk(value)
            return out
        if isinstance(node, list):
            return [_walk(item) for item in node]
        return node

    return _walk(payload)


def test_packaged_ontology_conforms_shacl():
    """
    Validate packaged ontology + shapes using OntologyManager SHACL validation.
    """
    om = OntologyManager()
    result = om.validate_ontology(inference="rdfs")
    assert bool(result.get("conforms", False)) is True


def test_ontology_manager_exports_nquads_without_packaged_context_file():
    context_path = _cf_ontology_source_path("ontology", "cf_context.nq")
    assert not context_path.exists()

    om = OntologyManager()
    exported = om.get_graph_nquads()
    assert CF_BASE in exported

    dataset = Dataset()
    dataset.parse(data=exported, format="nquads")
    total_triples = sum(len(graph) for graph in dataset.graphs())
    assert total_triples > 0


def test_pipeline_example_document_conforms_shacl(monkeypatch: pytest.MonkeyPatch):
    """
    Validate pipeline example RDF documents with SHACL.
    This validates semantics only; it does not execute pipelines.
    """
    monkeypatch.setenv("CF_ENABLE_STEP_PACKAGES", "0")
    ingest_rdf_documents(
        _repo_step_manifest_paths(),
        graph_type="step",
        package="repo-step-manifests",
    )
    om = OntologyManager()
    examples_dir = (
        _repo_root() / "sandcastle" / "cf_pipeline" / "cf_pipeline_engine" / "examples"
    )
    files = sorted(examples_dir.glob("*.nq"))
    assert files, "No pipeline examples found to validate."
    validated_count = 0

    for path in files:
        doc = load_rdf_document(path, label="test pipeline example")
        doc = _resolve_relative_context_paths(doc, path.parent)
        if not _contains_supported_pipeline_example(doc):
            continue
        if _contains_legacy_cogniflow_url(doc):
            continue
        result = om.validate_pipeline_document(doc)
        validated_count += 1
        assert bool(result.get("conforms", False)) is True, (
            f"{path.name} failed SHACL validation:\n{result.get('report_text', '')}"
        )
    if validated_count == 0:
        pytest.skip("No canonical pipeline examples without legacy-host URLs found.")


def test_combine_documents_merges_stub_nodes_with_full_definitions():
    merged = OntologyManager._combine_documents(
        [
            {
                "@context": {"cf": "https://cogniflow.odea-project.org/cf#"},
                "@graph": [
                    {
                        "@id": "cf:StepA",
                        "cf:hasParameter": {"@id": "cf:StepA_param_value"},
                    },
                    {"@id": "cf:StepA_param_value"},
                ],
            },
            {
                "@context": {"cf": "https://cogniflow.odea-project.org/cf#"},
                "@graph": [
                    {
                        "@id": "cf:StepA_param_value",
                        "cf:parameterName": "value",
                        "cf:defaultValue": 1,
                        "cf:valueContract": {"@id": "cf:StepA_param_value_contract"},
                    }
                ],
            },
        ]
    )

    graph = merged.get("@graph")
    assert isinstance(graph, list)
    by_id = {
        node.get("@id"): node
        for node in graph
        if isinstance(node, dict) and isinstance(node.get("@id"), str)
    }
    assert by_id["cf:StepA_param_value"]["cf:parameterName"] == "value"
    assert by_id["cf:StepA_param_value"]["cf:defaultValue"] == 1
    assert by_id["cf:StepA_param_value"]["cf:valueContract"] == {
        "@id": "cf:StepA_param_value_contract"
    }


def test_step_document_ingest_export_roundtrip_is_valid_document():
    """
    Ingest a step RDF document into quads, export it from DuckDB as N-Quads, and parse
    the export with RDFLib to verify syntactic/semantic validity.
    """
    basic_signal_steps = _resolve_steps_document(
        module_name="cf_basic_signal",
        repo_fallback=(
            _repo_root()
            / "sandcastle"
            / "cf_basic_steps"
            / "cf_basic_signal"
            / "src"
            / "cf_basic_signal"
            / "steps.nq"
        ),
    )
    basic_io_steps = _resolve_steps_document(
        module_name="cf_basic_io",
        repo_fallback=(
            _repo_root()
            / "sandcastle"
            / "cf_basic_steps"
            / "cf_basic_io"
            / "src"
            / "cf_basic_io"
            / "steps.nq"
        ),
    )
    ingest_rdf_documents([basic_signal_steps], graph_type="step", package="cf.basic.signal")
    ingest_rdf_documents([basic_io_steps], graph_type="step", package="cf.basic.io")
    basic_signal_doc = _load_json_document(basic_signal_steps)
    expected_average = _find_step_by_suffix(basic_signal_doc.get("@graph", []), "AverageStep")

    semantics_dir = resolve_semantics_dir()
    steps_db = resolve_semantics_db_path(semantics_dir, kind="steps")
    store = QuadStore(db_path=steps_db, semantics_dir=semantics_dir, read_only=True)
    exported = store.export_nquads()

    dataset = Dataset()
    dataset.parse(data=exported, format="nquads")
    total_triples = sum(len(graph) for graph in dataset.graphs())
    assert total_triples > 0


def test_pipeline_run_requires_session_link_in_shacl() -> None:
    shapes_path = (
        _repo_root()
        / "sandcastle"
        / "cf_ontology"
        / "src"
        / "cf_ontology"
        / "ontology"
        / "shapes"
        / "pipeline.shapes.nq"
    )
    shapes_graph = _load_graph_resource(shapes_path)

    missing_session_doc = {
        "@context": {
            "cf": CF_BASE,
            "ex": "http://example.org/",
            "xsd": "http://www.w3.org/2001/XMLSchema#",
        },
        "@graph": [
            {
                "@id": "ex:run_without_session",
                "@type": "cf:PipelineRun",
                "cf:runIndex": {"@type": "xsd:integer", "@value": 0},
            }
        ],
    }
    failing_graph = Graph().parse(data=json.dumps(missing_session_doc), format=_STRUCTURED_FORMAT)
    failing_conforms, _, _ = shacl_validate(
        failing_graph,
        shacl_graph=shapes_graph,
        inference="rdfs",
        abort_on_first=False,
        advanced=True,
    )
    assert bool(failing_conforms) is False

    valid_doc = {
        "@context": {
            "cf": CF_BASE,
            "ex": "http://example.org/",
            "xsd": "http://www.w3.org/2001/XMLSchema#",
        },
        "@graph": [
            {
                "@id": "ex:session_1",
                "@type": "cf:PipelineSession",
                "cf:sessionId": "session-1",
            },
            {
                "@id": "ex:run_with_session",
                "@type": "cf:PipelineRun",
                "cf:inSession": {"@id": "ex:session_1"},
                "cf:runIndex": {"@type": "xsd:integer", "@value": 1},
            },
        ],
    }
    passing_graph = Graph().parse(data=json.dumps(valid_doc), format=_STRUCTURED_FORMAT)
    passing_conforms, _, _ = shacl_validate(
        passing_graph,
        shacl_graph=shapes_graph,
        inference="rdfs",
        abort_on_first=False,
        advanced=True,
    )
    assert bool(passing_conforms) is True


def test_ontology_backed_datatype_normalization_uses_canonical_registry() -> None:
    assert normalize_parameter_type("cf:number") == "cf:double"
    assert normalize_parameter_type(f"{CF_BASE}float") == (
        f"{CF_BASE}double"
    )
    assert normalize_parameter_type("xsd:boolean") == "cf:boolean"
    assert parameter_type_to_literal_datatype("cf:number") == "xsd:double"
    assert parameter_type_to_literal_datatype(f"{CF_BASE}number") == (
        "http://www.w3.org/2001/XMLSchema#double"
    )
    assert normalize_port_literal_datatype("cf:number") == "xsd:double"
    assert normalize_port_literal_datatype(f"{CF_BASE}number") == (
        "http://www.w3.org/2001/XMLSchema#double"
    )
    assert normalize_port_literal_datatype("xsd:bolean") == "xsd:boolean"


def test_port_datatype_shape_accepts_registry_v0_canonical_set_only() -> None:
    shapes_path = (
        _repo_root()
        / "sandcastle"
        / "cf_ontology"
        / "src"
        / "cf_ontology"
        / "ontology"
        / "shapes"
        / "ports.shapes.nq"
    )
    shapes_graph = _load_graph_resource(shapes_path)

    context = {
        "cf": CF_BASE,
        "ex": "http://example.org/",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
    }

    for index, datatype in enumerate(supported_port_literal_datatypes()):
        valid_doc = {
            "@context": context,
            "@graph": [
                {
                    "@id": f"ex:port_{index}",
                    "@type": "cf:Port",
                    "cf:valueContract": {"@id": f"ex:port_{index}_contract"},
                },
                {
                    "@id": f"ex:port_{index}_contract",
                    "@type": "cf:ValueContract",
                    "cf:elementType": {"@id": datatype},
                    "cf:shapeKind": {"@id": "cf:shape_single"},
                }
            ],
        }
        valid_graph = Graph().parse(data=json.dumps(valid_doc), format=_STRUCTURED_FORMAT)
        conforms, _, _ = shacl_validate(
            valid_graph,
            shacl_graph=shapes_graph,
            inference="rdfs",
            abort_on_first=False,
            advanced=True,
        )
        assert bool(conforms) is True, f"Expected registry datatype {datatype} to pass SHACL."

    for datatype in ("xsd:dateTime", "cf:boolean"):
        invalid_doc = {
            "@context": context,
            "@graph": [
                {
                    "@id": "ex:invalid_port",
                    "@type": "cf:Port",
                    "cf:valueContract": {"@id": "ex:invalid_port_contract"},
                },
                {
                    "@id": "ex:invalid_port_contract",
                    "@type": "cf:ValueContract",
                    "cf:elementType": {"@id": datatype},
                    "cf:shapeKind": {"@id": "cf:shape_single"},
                }
            ],
        }
        invalid_graph = Graph().parse(data=json.dumps(invalid_doc), format=_STRUCTURED_FORMAT)
        conforms, _, _ = shacl_validate(
            invalid_graph,
            shacl_graph=shapes_graph,
            inference="rdfs",
            abort_on_first=False,
            advanced=True,
        )
        assert bool(conforms) is False, f"Expected non-canonical datatype {datatype} to fail SHACL."


def test_runtime_binding_vocab_matches_registry_supported_port_types() -> None:
    supported_types = set(supported_port_literal_datatypes())
    mapping_doc = _load_ontology_resource(("vocab", "datatypes_mapping.nq"))
    bindings_doc = _load_ontology_resource(("vocab", "runtime_storage_bindings.nq"))

    port_binding_rows = [
        node for node in mapping_doc.get("@graph", []) if _has_type(node, "cf:PortTypeBindingMapping")
    ]
    runtime_bindings = [node for node in bindings_doc.get("@graph", []) if _has_type(node, "cf:RuntimeBinding")]

    mapped_port_types = {node["cf:mappedPortType"]["@id"] for node in port_binding_rows}
    runtime_binding_types = {node["cf:forElementType"]["@id"] for node in runtime_bindings}
    runtime_binding_ids = {node["@id"] for node in runtime_bindings}
    referenced_runtime_binding_ids = {node["cf:mappedRuntimeBinding"]["@id"] for node in port_binding_rows}

    assert mapped_port_types == supported_types
    assert runtime_binding_types == supported_types
    assert referenced_runtime_binding_ids == runtime_binding_ids


def test_cf_ontology_local_registry_file_is_removed() -> None:
    assert not _cf_ontology_source_path("type_registry.v0.json").exists()
    source = _cf_ontology_source_path("datatypes_mapping.py").read_text(encoding="utf-8")
    assert "type_registry.v0.json" not in source


def test_storage_binding_vocab_keeps_storage_only_datetime_metadata() -> None:
    bindings_doc = _load_ontology_resource(("vocab", "runtime_storage_bindings.nq"))

    storage_bindings = [node for node in bindings_doc.get("@graph", []) if _has_type(node, "cf:StorageBinding")]
    storage_binding_types = {node["cf:forElementType"]["@id"] for node in storage_bindings}
    datetime_storage_binding = next(
        node for node in storage_bindings if node.get("@id") == "cf:storage_binding_xsd_dateTime_duckdb_v1"
    )

    assert "xsd:dateTime" in storage_binding_types
    assert datetime_storage_binding["cf:duckdbType"] == "TIMESTAMP"


def test_port_value_contract_shape_binding_and_table_schema_are_valid() -> None:
    shapes_path = (
        _repo_root()
        / "sandcastle"
        / "cf_ontology"
        / "src"
        / "cf_ontology"
        / "ontology"
        / "shapes"
        / "ports.shapes.nq"
    )
    shapes_graph = _load_graph_resource(shapes_path)

    context = {
        "cf": CF_BASE,
        "ex": "http://example.org/",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
    }

    valid_doc = {
        "@context": context,
        "@graph": [
            {
                "@id": "ex:runtime_binding_double",
                "@type": "cf:RuntimeBinding",
                "cf:forElementType": {"@id": "xsd:double"},
                "cf:runtimeKind": "DOUBLE",
                "cf:abiStruct": "double",
                "cf:typeVersion": {"@type": "xsd:integer", "@value": 1},
            },
            {
                "@id": "ex:storage_binding_double",
                "@type": "cf:StorageBinding",
                "cf:forElementType": {"@id": "xsd:double"},
                "cf:storageProfile": "native",
                "cf:duckdbType": "DOUBLE",
                "cf:typeVersion": {"@type": "xsd:integer", "@value": 1},
            },
            {
                "@id": "ex:vector_contract",
                "@type": "cf:ValueContract",
                "cf:elementType": {"@id": "xsd:double"},
                "cf:shapeKind": {"@id": "cf:shape_vector"},
                "cf:runtimeBinding": {"@id": "ex:runtime_binding_double"},
                "cf:storageBinding": {"@id": "ex:storage_binding_double"},
            },
            {
                "@id": "ex:table_column_value",
                "@type": "cf:TableColumnSpec",
                "cf:columnName": "value",
                "cf:columnElementType": {"@id": "xsd:double"},
                "cf:columnNullable": {"@type": "xsd:boolean", "@value": False},
            },
            {
                "@id": "ex:table_schema_v1",
                "@type": "cf:TableSchema",
                "cf:hasColumn": {"@id": "ex:table_column_value"},
            },
            {
                "@id": "ex:table_contract",
                "@type": "cf:ValueContract",
                "cf:elementType": {"@id": "xsd:double"},
                "cf:shapeKind": {"@id": "cf:shape_table"},
                "cf:tableSchema": {"@id": "ex:table_schema_v1"},
            },
            {
                "@id": "ex:vector_port",
                "@type": "cf:Port",
                "cf:valueContract": {"@id": "ex:vector_contract"},
            },
            {
                "@id": "ex:table_port",
                "@type": "cf:Port",
                "cf:valueContract": {"@id": "ex:table_contract"},
            },
        ],
    }

    valid_graph = Graph().parse(data=json.dumps(valid_doc), format=_STRUCTURED_FORMAT)
    conforms, _, _ = shacl_validate(
        valid_graph,
        shacl_graph=shapes_graph,
        inference="rdfs",
        abort_on_first=False,
        advanced=True,
    )
    assert bool(conforms) is True


def test_port_value_contract_shape_binding_and_table_schema_reject_invalid() -> None:
    shapes_path = (
        _repo_root()
        / "sandcastle"
        / "cf_ontology"
        / "src"
        / "cf_ontology"
        / "ontology"
        / "shapes"
        / "ports.shapes.nq"
    )
    shapes_graph = _load_graph_resource(shapes_path)

    context = {
        "cf": CF_BASE,
        "ex": "http://example.org/",
        "xsd": "http://www.w3.org/2001/XMLSchema#",
    }

    invalid_docs = [
        {
            "@context": context,
            "@graph": [
                {
                    "@id": "ex:invalid_contract_shape",
                    "@type": "cf:ValueContract",
                    "cf:elementType": {"@id": "xsd:double"},
                    "cf:shapeKind": {"@id": "cf:shape_cube"},
                },
                {
                    "@id": "ex:invalid_port_shape",
                    "@type": "cf:Port",
                    "cf:valueContract": {"@id": "ex:invalid_contract_shape"},
                },
            ],
        },
        {
            "@context": context,
            "@graph": [
                {
                    "@id": "ex:invalid_contract_table",
                    "@type": "cf:ValueContract",
                    "cf:elementType": {"@id": "xsd:double"},
                    "cf:shapeKind": {"@id": "cf:shape_table"},
                },
                {
                    "@id": "ex:invalid_port_table",
                    "@type": "cf:Port",
                    "cf:valueContract": {"@id": "ex:invalid_contract_table"},
                },
            ],
        },
        {
            "@context": context,
            "@graph": [
                {
                    "@id": "ex:invalid_runtime_binding",
                    "@type": "cf:RuntimeBinding",
                    "cf:forElementType": {"@id": "xsd:double"},
                    "cf:runtimeKind": "DOUBLE",
                    "cf:typeVersion": {"@type": "xsd:integer", "@value": 1},
                }
            ],
        },
        {
            "@context": context,
            "@graph": [
                {
                    "@id": "ex:invalid_table_column",
                    "@type": "cf:TableColumnSpec",
                    "cf:columnName": "ts",
                    "cf:columnElementType": {"@id": "xsd:duration"},
                    "cf:columnNullable": {"@type": "xsd:boolean", "@value": False},
                },
                {
                    "@id": "ex:invalid_table_schema",
                    "@type": "cf:TableSchema",
                    "cf:hasColumn": {"@id": "ex:invalid_table_column"},
                },
                {
                    "@id": "ex:invalid_contract_table_col",
                    "@type": "cf:ValueContract",
                    "cf:elementType": {"@id": "xsd:double"},
                    "cf:shapeKind": {"@id": "cf:shape_table"},
                    "cf:tableSchema": {"@id": "ex:invalid_table_schema"},
                },
            ],
        },
    ]

    for invalid_doc in invalid_docs:
        invalid_graph = Graph().parse(data=json.dumps(invalid_doc), format=_STRUCTURED_FORMAT)
        conforms, _, _ = shacl_validate(
            invalid_graph,
            shacl_graph=shapes_graph,
            inference="rdfs",
            abort_on_first=False,
            advanced=True,
        )
        assert bool(conforms) is False


def test_step_package_value_contract_adoption_conforms_shacl() -> None:
    shapes_path = (
        _repo_root()
        / "sandcastle"
        / "cf_ontology"
        / "src"
        / "cf_ontology"
        / "ontology"
        / "shapes"
        / "ports.shapes.nq"
    )
    shapes_graph = _load_graph_resource(shapes_path)

    step_paths = [
        _repo_root()
        / "sandcastle"
        / "cf_basic_steps"
        / "cf_basic_dev"
        / "src"
        / "cf_basic_dev"
        / "steps.nq",
        _repo_root()
        / "sandcastle"
        / "cf_basic_steps"
        / "cf_basic_io"
        / "src"
        / "cf_basic_io"
        / "steps.nq",
        _repo_root()
        / "sandcastle"
        / "cf_basic_steps"
        / "cf_basic_signal"
        / "src"
        / "cf_basic_signal"
        / "steps.nq",
        _repo_root()
        / "sandcastle"
        / "cf_basic_steps"
        / "cf_basic_sinks"
        / "src"
        / "cf_basic_sinks"
        / "steps.nq",
    ]
    expected_vector_contracts_by_package = {
        "cf_basic_dev": {
            "basic-dev/InlineSource_output_data": ("xsd:double", "cf:shape_vector"),
        },
        "cf_basic_io": {
            "basic-io/HandleSinkStep_input_data": ("xsd:double", "cf:shape_vector"),
        },
        "cf_basic_signal": {
            "basic-signal/AverageStep_input_values": ("xsd:double", "cf:shape_vector"),
            "basic-signal/FifoWindowBufferStep_output_values": ("xsd:double", "cf:shape_vector"),
        },
    }

    for step_path in step_paths:
        doc = _load_json_document(step_path)
        graph_nodes = doc.get("@graph")
        assert isinstance(graph_nodes, list)
        by_id = {node.get("@id"): node for node in graph_nodes if isinstance(node, dict)}
        expected_vector_contracts = expected_vector_contracts_by_package.get(step_path.parent.name, {})
        observed_vector_contracts: set[str] = set()

        typed_ports = [
            node
            for node in graph_nodes
            if isinstance(node, dict)
            and (
                _has_type(node, "cf:InputPort")
                or _has_type(node, "cf:OutputPort")
                or _has_type(node, "cf:Port")
            )
            and isinstance(node.get("cf:valueContract"), dict)
            and isinstance(node["cf:valueContract"].get("@id"), str)
        ]
        assert typed_ports, f"No contracted ports found in {step_path.name}."

        for port in typed_ports:
            contract_ref = port.get("cf:valueContract")
            assert isinstance(contract_ref, dict) and isinstance(contract_ref.get("@id"), str), (
                f"Missing valueContract on port {port.get('@id')} in {step_path.name}."
            )
            contract = by_id.get(contract_ref["@id"])
            assert isinstance(contract, dict)
            assert _has_type(contract, "cf:ValueContract")
            port_id = port.get("@id")
            matched_suffix = next(
                (suffix for suffix in expected_vector_contracts if _id_has_suffix(port_id, suffix)),
                None,
            )
            if matched_suffix:
                expected_element_type, expected_shape = expected_vector_contracts[matched_suffix]
                observed_vector_contracts.add(matched_suffix)
                assert contract.get("cf:elementType", {}).get("@id") == expected_element_type
                assert contract.get("cf:shapeKind", {}).get("@id") == expected_shape
            else:
                assert isinstance(contract.get("cf:elementType", {}).get("@id"), str)
                assert contract.get("cf:shapeKind", {}).get("@id") in {
                    "cf:shape_single",
                    "cf:shape_vector",
                    "cf:shape_matrix",
                    "cf:shape_table",
                }

        assert observed_vector_contracts == set(expected_vector_contracts), (
            f"Missing selected migrated vector ports in {step_path.name}: "
            f"{set(expected_vector_contracts) - observed_vector_contracts}"
        )

        selected_graph_nodes: list[dict] = []
        selected_ids: set[str] = set()
        for port in typed_ports:
            port_id = port.get("@id")
            if isinstance(port_id, str) and port_id not in selected_ids:
                port_copy = deepcopy(port)
                # Isolate contract adoption checks from legacy uncertainty typing noise.
                port_copy.pop("cf:uncertaintyFor", None)
                selected_graph_nodes.append(port_copy)
                selected_ids.add(port_id)

            contract_ref = port.get("cf:valueContract", {})
            contract_id = contract_ref.get("@id") if isinstance(contract_ref, dict) else None
            if isinstance(contract_id, str) and contract_id not in selected_ids:
                contract = by_id.get(contract_id)
                if isinstance(contract, dict):
                    selected_graph_nodes.append(deepcopy(contract))
                    selected_ids.add(contract_id)

        reduced_doc = {"@context": doc.get("@context"), "@graph": selected_graph_nodes}
        resolved_doc = _resolve_relative_context_paths(reduced_doc, step_path.parent)
        data_graph = Graph().parse(data=json.dumps(resolved_doc), format=_STRUCTURED_FORMAT)
        conforms, _, _ = shacl_validate(
            data_graph,
            shacl_graph=shapes_graph,
            inference="rdfs",
            abort_on_first=False,
            advanced=True,
        )
        assert bool(conforms) is True, f"Adopted step manifest failed SHACL: {step_path.name}"


def test_pipeline_examples_reference_step_manifests_with_value_contracts() -> None:
    example_paths = [
        _repo_root()
        / "sandcastle"
        / "cf_pipeline"
        / "cf_pipeline_engine"
        / "examples"
        / "omp_reduction_parallel.nq",
        _repo_root()
        / "sandcastle"
        / "cf_pipeline"
        / "cf_pipeline_engine"
        / "examples"
        / "opcua_fifo_avg_to_duckdb_parquet_triggered.nq",
        _repo_root()
        / "sandcastle"
        / "cf_pipeline"
        / "cf_pipeline_engine"
        / "examples"
        / "opcua_temp_fifo_avg_to_duckdb_parquet_triggered.nq",
    ]

    for example_path in example_paths:
        doc = load_rdf_document(example_path, label="pipeline example")
        resolved = _resolve_relative_context_paths(doc, example_path.parent)

        steps_paths: list[str] = []
        if isinstance(resolved, dict) and "@graph" in resolved:
            graph_nodes = resolved.get("@graph")
            if isinstance(graph_nodes, list):
                for node in graph_nodes:
                    if isinstance(node, dict) and node.get("@type") == "cf:StepsHeader":
                        _extend_string_values(steps_paths, node.get("cf:stepsPath", []))
        elif isinstance(resolved, dict):
            header = resolved.get("cf:hasStepsHeader", {})
            if isinstance(header, dict):
                _extend_string_values(steps_paths, header.get("cf:stepsPath", []))

        assert steps_paths, f"No step manifests referenced by {example_path.name}."
        for step_path_text in steps_paths:
            step_path = Path(step_path_text)
            if not step_path.is_absolute():
                step_path = (example_path.parent / step_path).resolve()
            assert step_path.is_file(), f"Missing referenced steps manifest: {step_path}"
            step_doc = _load_json_document(step_path)
            step_graph = step_doc.get("@graph", [])
            typed_ports = [
                node
                for node in step_graph
                if isinstance(node, dict)
                and (
                    _has_type(node, "cf:InputPort")
                    or _has_type(node, "cf:OutputPort")
                    or _has_type(node, "cf:Port")
                )
                and isinstance(node.get("cf:valueContract"), dict)
                and isinstance(node["cf:valueContract"].get("@id"), str)
            ]
            assert typed_ports, f"No contracted ports in referenced manifest {step_path.name}."
            assert all("cf:valueContract" in port for port in typed_ports), (
                f"Referenced manifest lacks valueContract adoption: {step_path.name}"
            )


def test_step_package_value_contract_uses_supported_shape_kind() -> None:
    step_path = (
        _repo_root()
        / "sandcastle"
        / "cf_basic_steps"
        / "cf_basic_dev"
        / "src"
        / "cf_basic_dev"
        / "steps.nq"
    )
    doc = _load_json_document(step_path)
    graph_nodes = doc.get("@graph")
    assert isinstance(graph_nodes, list)

    by_id = {node.get("@id"): node for node in graph_nodes if isinstance(node, dict)}
    typed_port = next(
        node
        for node in graph_nodes
        if isinstance(node, dict)
        and (_has_type(node, "cf:InputPort") or _has_type(node, "cf:OutputPort") or _has_type(node, "cf:Port"))
        and isinstance(node.get("cf:valueContract"), dict)
        and isinstance(node["cf:valueContract"].get("@id"), str)
    )
    contract_ref = typed_port.get("cf:valueContract", {})
    assert isinstance(contract_ref, dict) and isinstance(contract_ref.get("@id"), str)
    contract_node = by_id.get(contract_ref["@id"])
    assert isinstance(contract_node, dict)
    shape_kind = contract_node.get("cf:shapeKind")
    assert isinstance(shape_kind, dict)
    assert shape_kind.get("@id") in {
        "cf:shape_single",
        "cf:shape_vector",
        "cf:shape_matrix",
        "cf:shape_table",
    }



"""Centralized datatype mapping helpers for ontology + web pipelines."""

from __future__ import annotations

from functools import lru_cache
import json
from pathlib import Path
from typing import Final

from rdflib import Dataset

from .namespaces import CF_BASE as _CF_BASE
from .namespaces import CFRUN_BASE as _CFRUN_BASE
from .namespaces import canonical_context

_ONTOLOGY_DIRNAME: Final = "ontology"
_DATATYPES_MAPPING_FILENAME: Final = "datatypes_mapping.nq"
_STRUCTURED_FORMAT: Final = "json" + "-ld"
_EXTRA_ALIAS_COMPACTS: Final = {
    "boolean": ("xsd:bolean",),
    "double": ("cf:float", "cf:number"),
}


def _normalize_input(value: str | None) -> str | None:
    if not value:
        return None
    text = value.strip()
    return text or None


def _prefers_iri(value: str) -> bool:
    lowered = value.lower()
    return lowered.startswith("http://") or lowered.startswith("https://")


@lru_cache(maxsize=1)
def _prefix_context() -> dict[str, str]:
    context = canonical_context()
    context["cf"] = _CF_BASE
    context["cfrun"] = _CFRUN_BASE
    return {
        key: value
        for key, value in context.items()
        if isinstance(key, str) and isinstance(value, str)
    }


def _document_from_dataset(dataset: Dataset) -> dict:
    serialized = dataset.serialize(
        format=_STRUCTURED_FORMAT,
        context=_prefix_context(),
        auto_compact=True,
    )
    if isinstance(serialized, bytes):
        serialized = serialized.decode("utf-8")
    payload = json.loads(serialized)
    if isinstance(payload, list):
        return {"@context": _prefix_context(), "@graph": payload}
    if isinstance(payload, dict):
        if "@graph" in payload:
            if "@context" not in payload:
                payload["@context"] = _prefix_context()
            return payload
        return {"@context": _prefix_context(), "@graph": [payload]}
    raise ValueError("Unsupported RDF document payload generated from N-Quads dataset.")


def _expand_identifier(value: str) -> str:
    if _prefers_iri(value):
        return value
    prefix, separator, suffix = value.partition(":")
    if not separator:
        return value
    base = _prefix_context().get(prefix)
    if not base:
        return value
    return f"{base}{suffix}"


def _identifier_variants(compact_value: str) -> set[str]:
    variants = {compact_value}
    expanded = _expand_identifier(compact_value)
    variants.add(expanded)
    return variants


def _has_type(node: dict[str, object], expected: str) -> bool:
    node_type = node.get("@type")
    if isinstance(node_type, str):
        return node_type == expected
    if isinstance(node_type, list):
        return expected in node_type
    return False


@lru_cache(maxsize=1)
def _registry() -> dict:
    path = Path(__file__).parent / _ONTOLOGY_DIRNAME / "vocab" / _DATATYPES_MAPPING_FILENAME
    dataset = Dataset()
    dataset.parse(str(path), format="nquads")
    payload = _document_from_dataset(dataset)
    graph = payload.get("@graph", [])

    parameter_aliases: dict[str, dict] = {}
    port_aliases: dict[str, dict] = {}
    supported_port_types_compact: list[str] = []
    supported_port_types_iri: list[str] = []

    parameter_rows = [node for node in graph if isinstance(node, dict) and _has_type(node, "cf:ParameterTypeDatatypeMappingShape")]
    port_rows = [node for node in graph if isinstance(node, dict) and _has_type(node, "cf:PortTypeBindingMapping")]
    port_rows_by_compact = {
        str(node["cf:mappedPortType"]["@id"]): node
        for node in port_rows
        if isinstance(node.get("cf:mappedPortType"), dict) and "@id" in node["cf:mappedPortType"]
    }

    for node in port_rows:
        mapped_port_type = node.get("cf:mappedPortType")
        if not isinstance(mapped_port_type, dict):
            continue
        compact_port_type = mapped_port_type.get("@id")
        if not isinstance(compact_port_type, str):
            continue
        supported_port_types_compact.append(compact_port_type)
        supported_port_types_iri.append(_expand_identifier(compact_port_type))

    datatypes: list[dict[str, object]] = []
    for node in parameter_rows:
        mapped_parameter_type = node.get("cf:mappedParameterType")
        mapped_literal_datatype = node.get("cf:mappedLiteralDatatype")
        if not isinstance(mapped_parameter_type, dict) or not isinstance(mapped_literal_datatype, dict):
            continue

        parameter_compact = mapped_parameter_type.get("@id")
        port_compact = mapped_literal_datatype.get("@id")
        if not isinstance(parameter_compact, str) or not isinstance(port_compact, str):
            continue

        entry_id = parameter_compact.split(":", 1)[1]
        entry = {
            "id": entry_id,
            "parameter": {
                "canonical_compact": parameter_compact,
                "canonical_iri": _expand_identifier(parameter_compact),
            },
            "port": {
                "canonical_compact": port_compact,
                "canonical_iri": _expand_identifier(port_compact),
            },
        }
        datatypes.append(entry)

        aliases = set()
        aliases.update(_identifier_variants(parameter_compact))
        aliases.update(_identifier_variants(port_compact))
        for extra_compact in _EXTRA_ALIAS_COMPACTS.get(entry_id, ()):
            aliases.update(_identifier_variants(extra_compact))

        for alias in aliases:
            parameter_aliases[alias] = entry
            port_aliases[alias] = entry

    payload["datatypes"] = datatypes
    payload["_port_binding_rows"] = tuple(sorted(port_rows_by_compact))

    payload["_parameter_aliases"] = parameter_aliases
    payload["_port_aliases"] = port_aliases
    payload["_supported_port_types_compact"] = tuple(supported_port_types_compact)
    payload["_supported_port_types_iri"] = tuple(supported_port_types_iri)
    return payload


def _canonical_value(section: dict[str, object], prefer_iri: bool) -> str:
    key = "canonical_iri" if prefer_iri else "canonical_compact"
    return str(section[key])


def normalize_parameter_type(value: str | None) -> str | None:
    text = _normalize_input(value)
    if not text:
        return None
    entry = _registry()["_parameter_aliases"].get(text)
    if not entry:
        return text
    return _canonical_value(entry["parameter"], prefer_iri=_prefers_iri(text))


def normalize_literal_datatype(value: str | None) -> str | None:
    text = _normalize_input(value)
    if not text:
        return None
    entry = _registry()["_port_aliases"].get(text)
    if not entry:
        return text
    return _canonical_value(entry["port"], prefer_iri=_prefers_iri(text))


def parameter_type_to_literal_datatype(parameter_type: str | None) -> str | None:
    text = _normalize_input(parameter_type)
    if not text:
        return None
    entry = _registry()["_parameter_aliases"].get(text)
    if not entry:
        return None
    return _canonical_value(entry["port"], prefer_iri=_prefers_iri(text))


def normalize_port_literal_datatype(port_type: str | None) -> str | None:
    text = _normalize_input(port_type)
    if not text:
        return None
    entry = _registry()["_port_aliases"].get(text)
    if entry:
        return _canonical_value(entry["port"], prefer_iri=_prefers_iri(text))
    mapped = parameter_type_to_literal_datatype(text)
    if mapped:
        return mapped
    return text


def supported_port_literal_datatypes(*, iri: bool = False) -> tuple[str, ...]:
    key = "_supported_port_types_iri" if iri else "_supported_port_types_compact"
    return _registry()[key]




from __future__ import annotations

from pathlib import Path

from cf_ontology import pipeline_state_runtime as runtime


class _Proc:
    def __init__(self, returncode: int = 0) -> None:
        self.returncode = returncode


def test_run_engine_does_not_set_direct_override_env_by_default(monkeypatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run(cmd, check=False, env=None):  # noqa: ANN001, ANN202
        captured["cmd"] = cmd
        captured["check"] = check
        captured["env"] = env
        return _Proc(0)

    monkeypatch.setattr(runtime.subprocess, "run", _fake_run)
    code = runtime._run_engine(
        Path("demo.nq"),
        steps_paths=[],
        plugin_paths=[],
        exe="cf_pipeline_v2",
    )

    assert code == 0
    assert captured["check"] is False
    assert captured["env"] is None


def test_run_engine_applies_explicit_env_override(monkeypatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run(cmd, check=False, env=None):  # noqa: ANN001, ANN202
        captured["cmd"] = cmd
        captured["check"] = check
        captured["env"] = env
        return _Proc(0)

    monkeypatch.setattr(runtime.subprocess, "run", _fake_run)
    code = runtime._run_engine(
        Path("demo.nq"),
        steps_paths=[],
        plugin_paths=[],
        exe="cf_pipeline_v2",
        env_overrides={"CF_ALLOW_DIRECT_ENGINE": "1"},
    )

    assert code == 0
    assert captured["check"] is False
    assert isinstance(captured["env"], dict)
    assert captured["env"]["CF_ALLOW_DIRECT_ENGINE"] == "1"


def test_run_engine_prepends_plugin_dirs_to_path(monkeypatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run(cmd, check=False, env=None):  # noqa: ANN001, ANN202
        captured["cmd"] = cmd
        captured["check"] = check
        captured["env"] = env
        return _Proc(0)

    monkeypatch.setenv("PATH", "BASE_PATH")
    monkeypatch.setattr(runtime.subprocess, "run", _fake_run)
    code = runtime._run_engine(
        Path("demo.nq"),
        steps_paths=[],
        plugin_paths=["C:/plugins/a", "C:/plugins/b", "C:/plugins/a"],
        exe="cf_pipeline_v2",
    )

    assert code == 0
    assert captured["check"] is False
    assert isinstance(captured["env"], dict)
    expected_path = "C:/plugins/a" + runtime.os.pathsep + "C:/plugins/b" + runtime.os.pathsep + "BASE_PATH"
    assert captured["env"]["PATH"] == expected_path


def test_run_engine_uses_resolved_engine_when_exe_missing(monkeypatch) -> None:
    captured: dict[str, object] = {}

    def _fake_run(cmd, check=False, env=None):  # noqa: ANN001, ANN202
        captured["cmd"] = cmd
        captured["check"] = check
        captured["env"] = env
        return _Proc(0)

    monkeypatch.setattr(runtime, "_find_pipeline_exe", lambda: "resolved-engine")
    monkeypatch.setattr(runtime.subprocess, "run", _fake_run)
    code = runtime._run_engine(
        Path("demo.nq"),
        steps_paths=[],
        plugin_paths=[],
        exe=None,
    )

    assert code == 0
    assert isinstance(captured["cmd"], list)
    assert captured["cmd"][0] == "resolved-engine"



"""Helpers for loading RDF documents from structured JSON or N-Quads inputs."""

from __future__ import annotations

from pathlib import Path
from typing import Any, Dict, Optional, Tuple, TypeAlias
import json

from rdflib import Dataset

from .namespaces import canonical_context


DocumentInput: TypeAlias = Dict[str, Any] | str | Path
_STRUCTURED_FORMAT = "json" + "-ld"


def canonical_json(value: Any) -> str:
    return json.dumps(value, sort_keys=True, separators=(",", ":"), ensure_ascii=False)


def _ensure_document_dict(payload: Any, *, source: str) -> Dict[str, Any]:
    if isinstance(payload, dict):
        return payload
    if isinstance(payload, list):
        return {
            "@context": canonical_context(),
            "@graph": payload,
        }
    raise ValueError(f"{source} must decode to a JSON object or graph list")


def _load_json_document(text: str, *, source: str) -> Dict[str, Any]:
    try:
        payload = json.loads(text)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        raise ValueError(f"Failed to parse {source}: {exc}") from exc
    return _ensure_document_dict(payload, source=source)


def _load_json_payload(path: Path) -> Dict[str, Any]:
    payload = json.loads(path.read_text(encoding="utf-8"))

    def _rewrite_context(value: Any) -> Any:
        if isinstance(value, list):
            return [_rewrite_context(item) for item in value]
        if isinstance(value, dict):
            return {key: _rewrite_context(item) for key, item in value.items()}
        if isinstance(value, str):
            candidate = (path.parent / value).resolve() if not value.startswith(("http://", "https://", "file://")) else None
            if candidate is not None and candidate.exists():
                loaded = json.loads(candidate.read_text(encoding="utf-8"))
                if isinstance(loaded, dict) and "@context" in loaded:
                    return _rewrite_context(loaded["@context"])
                return _rewrite_context(loaded)
            if value.replace("\\", "/").endswith("cf_context." + "json" + "ld"):
                return canonical_context()
        return value

    if isinstance(payload, dict) and "@context" in payload:
        payload["@context"] = _rewrite_context(payload["@context"])
    return _ensure_document_dict(payload, source=str(path))


def _dataset_to_document(dataset: Dataset, *, source: str) -> Dict[str, Any]:
    payload = dataset.serialize(
        format=_STRUCTURED_FORMAT,
        context=canonical_context(),
        auto_compact=True,
    )
    if isinstance(payload, bytes):
        payload = payload.decode("utf-8")
    document = _load_json_document(payload, source=source)
    if "@context" not in document:
        document["@context"] = canonical_context()
    return document


def _load_nquads_document(text: str, *, source: str) -> Dict[str, Any]:
    dataset = Dataset()
    try:
        dataset.parse(data=text, format="nquads")
    except Exception as exc:  # pylint: disable=broad-exception-caught
        raise ValueError(f"Failed to parse {source}: {exc}") from exc
    return _dataset_to_document(dataset, source=source)


def _load_json_path_document(path: Path, *, source: str) -> Dict[str, Any]:
    payload = _load_json_payload(path)
    dataset = Dataset()
    try:
        dataset.parse(data=canonical_json(payload), format=_STRUCTURED_FORMAT, publicID=path.as_uri())
    except Exception as exc:  # pylint: disable=broad-exception-caught
        raise ValueError(f"Failed to parse {source}: {exc}") from exc
    return _dataset_to_document(dataset, source=source)


def _load_path_document(path: Path, *, label: str) -> Tuple[Dict[str, Any], Optional[Path], str, str]:
    resolved = path.expanduser().resolve()
    if not resolved.is_file():
        raise FileNotFoundError(f"{label.capitalize()} file not found: {resolved}")
    source = f"{label} at {resolved}"
    suffix = resolved.suffix.lower()
    if suffix in {"." + "json" + "ld", ".json"}:
        document = _load_json_path_document(resolved, source=source)
    elif suffix in {".nq", ".nquads"}:
        text = resolved.read_text(encoding="utf-8")
        document = _load_nquads_document(text, source=source)
    else:
        raise ValueError(
            f"Unsupported {label} format at {resolved}: expected .json, .nq, or .nquads"
        )
    return document, resolved.parent, canonical_json(document), str(resolved)


def load_rdf_document_input(
    value: DocumentInput,
    *,
    label: str = "RDF document",
) -> Tuple[Dict[str, Any], Optional[Path], str, Optional[str]]:
    if isinstance(value, dict):
        return value, None, canonical_json(value), None
    if isinstance(value, Path):
        return _load_path_document(value, label=label)
    if isinstance(value, str):
        candidate = Path(value)
        if candidate.is_file():
            return _load_path_document(candidate, label=label)
        stripped = value.lstrip()
        source = label
        if stripped.startswith("{") or stripped.startswith("["):
            document = _load_json_document(value, source=source)
        else:
            document = _load_nquads_document(value, source=source)
        return document, None, canonical_json(document), None
    raise TypeError(f"Unsupported {label} type: {type(value)!r}")


def load_rdf_document(value: DocumentInput, *, label: str = "RDF document") -> Dict[str, Any]:
    document, _, _, _ = load_rdf_document_input(value, label=label)
    return document


from __future__ import annotations

import json
from pathlib import Path

from rdflib import Dataset

from cf_ontology.namespaces import CF_BASE
from cf_ontology.pipeline_id import derive_pipeline_id_from_document
from cf_ontology.quad_store import QuadStore

_STRUCTURED_FORMAT = "json" + "-ld"


def _pipeline_doc() -> dict:
    return {
        "@context": {
            "cf": CF_BASE,
            "ex": "http://example.org/",
        },
        "@graph": [
            {
                "@id": "ex:pipe-a",
                "@type": "cf:ProcessingPipeline",
                "cf:hasPipelineNode": {"@id": "ex:n1"},
            },
            {
                "@id": "ex:n1",
                "@type": "cf:PipelineNode",
                "cf:hasParameterBinding": {"@id": "ex:n1_endpoint_binding"},
            },
            {
                "@id": "ex:n1_endpoint_binding",
                "@type": "cf:ParameterBinding",
                "cf:bindsParameter": {"@id": "ex:InputStep_param_endpoint"},
                "cf:value": "opc.tcp://127.0.0.1:4840/VirtualPhServer",
            },
        ],
    }


def _write_nquads(path: Path, document: dict) -> Path:
    dataset = Dataset()
    dataset.parse(data=json.dumps(document), format=_STRUCTURED_FORMAT)
    payload = dataset.serialize(format="nquads")
    if isinstance(payload, bytes):
        payload = payload.decode("utf-8")
    path.write_text(payload, encoding="utf-8")
    return path


def test_derive_pipeline_id_from_nquads_document(tmp_path: Path) -> None:
    path = _write_nquads(tmp_path / "pipeline.nq", _pipeline_doc())

    assert derive_pipeline_id_from_document(path) == "pipe-a"


def test_commit_pipeline_revision_accepts_nquads_document(tmp_path: Path) -> None:
    semantics_dir = tmp_path / "semantics"
    store = QuadStore(
        db_path=semantics_dir / "cf-pipelines.duckdb",
        semantics_dir=semantics_dir,
    )
    store.init_schema(create_pipeline_tables=True)
    path = _write_nquads(tmp_path / "pipeline.nq", _pipeline_doc())

    rev = store.commit_pipeline_revision("pipe-a", path, created_by="alice", message="nq")

    assert rev == 1
    exported = store.load_pipeline_revision_nquads("pipe-a", rev)
    assert "https://cogniflow.odea-project.org/cf#ProcessingPipeline" in exported
    assert "opc.tcp://127.0.0.1:4840/VirtualPhServer" in exported


from __future__ import annotations

from copy import deepcopy

import pytest

from cf_ontology.namespaces import CF_BASE
from cf_ontology.pipeline_state_runtime import _apply_invocation_overrides


class _FakeOntologyManager:
    def __init__(self, *, conforms: bool = True, report_text: str = "") -> None:
        self._conforms = conforms
        self._report_text = report_text
        self.last_doc = None

    def validate_pipeline_document(self, doc):  # noqa: ANN001
        self.last_doc = doc
        return {
            "conforms": self._conforms,
            "report_text": self._report_text,
        }


def _pipeline_doc() -> dict:
    return {
        "@context": {
            "cf": CF_BASE,
            "ex": "http://example.org/",
            "schema": "https://schema.org/",
        },
        "@graph": [
            {
                "@id": "ex:pipeline",
                "@type": "cf:ProcessingPipeline",
                "cf:hasPipelineNode": {"@id": "ex:n1"},
            },
            {
                "@id": "ex:n1",
                "@type": "cf:PipelineNode",
                "cf:nodeDefinition": {"@id": "ex:InputStep"},
                "cf:hasParameterBinding": {"@id": "ex:n1_endpoint_binding"},
            },
            {
                "@id": "ex:InputStep",
                "@type": "cf:ProcessingStep",
                "cf:stepCategory": {"@id": "cf:input"},
            },
            {
                "@id": "ex:n1_endpoint_binding",
                "@type": "cf:ParameterBinding",
                "cf:bindsParameter": {"@id": "ex:InputStep_param_endpoint"},
                "cf:value": "opc.tcp://127.0.0.1:4840/VirtualPhServer",
            },
        ],
    }


def _invocation_doc(*, key: str = "endpoint", value: str = "opc.tcp://127.0.0.1:4841/VirtualPhServer") -> dict:
    return {
        "@context": {
            "cf": CF_BASE,
            "ex": "http://example.org/",
            "schema": "https://schema.org/",
        },
        "@graph": [
            {
                "@id": "ex:invocation",
                "@type": "cf:PipelineInvocation",
                "cf:invokesPipeline": {"@id": "ex:pipeline"},
                "cf:overridesEntryPoint": {"@id": "ex:override_n1"},
            },
            {
                "@id": "ex:override_n1",
                "@type": "cf:EntryPointOverride",
                "cf:targetsNode": {"@id": "ex:n1"},
                "cf:overrideSettings": {"@id": "ex:override_settings_n1"},
            },
            {
                "@id": "ex:override_settings_n1",
                "@type": "schema:StructuredValue",
                "schema:additionalProperty": {"@id": "ex:override_settings_n1_property"},
            },
            {
                "@id": "ex:override_settings_n1_property",
                "@type": "schema:PropertyValue",
                "schema:propertyID": key,
                "schema:value": value,
            },
        ],
    }


def _binding_value(pipeline_doc: dict) -> str:
    for node in pipeline_doc.get("@graph", []):
        if node.get("@id") == "ex:n1_endpoint_binding":
            return str(node.get("cf:value"))
    raise AssertionError("ex:n1_endpoint_binding missing")


def test_apply_invocation_overrides_updates_input_binding() -> None:
    pipeline_doc = _pipeline_doc()
    pipeline_before = deepcopy(pipeline_doc)
    invocation_doc = _invocation_doc()
    fake_om = _FakeOntologyManager(conforms=True)

    patched = _apply_invocation_overrides(fake_om, pipeline_doc, invocation_doc)

    assert _binding_value(pipeline_before) == "opc.tcp://127.0.0.1:4840/VirtualPhServer"
    assert _binding_value(patched) == "opc.tcp://127.0.0.1:4841/VirtualPhServer"
    assert _binding_value(pipeline_doc) == "opc.tcp://127.0.0.1:4840/VirtualPhServer"
    assert fake_om.last_doc is not None


def test_apply_invocation_overrides_rejects_unknown_parameter_key() -> None:
    pipeline_doc = _pipeline_doc()
    invocation_doc = _invocation_doc(key="unknown_setting")
    fake_om = _FakeOntologyManager(conforms=True)

    with pytest.raises(ValueError, match="does not match any parameter binding"):
        _apply_invocation_overrides(fake_om, pipeline_doc, invocation_doc)


def test_apply_invocation_overrides_requires_conforming_invocation() -> None:
    pipeline_doc = _pipeline_doc()
    invocation_doc = _invocation_doc()
    fake_om = _FakeOntologyManager(conforms=False, report_text="validation failed")

    with pytest.raises(ValueError, match="Invocation SHACL validation failed"):
        _apply_invocation_overrides(fake_om, pipeline_doc, invocation_doc)



"""DuckDB-backed runtime state store for pipeline execution."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
import json
from pathlib import Path
from typing import Any, Optional
import uuid

import duckdb


_LOCAL_TIMEZONE = datetime.now().astimezone().tzinfo or timezone.utc

@dataclass(frozen=True)
class PendingRunEvent:
    event_id: str
    pipeline_id: str
    event_type: str
    source: str
    payload: Any
    created_at: Optional[datetime]
    dedupe_key: Optional[str]
    source_event_id: Optional[str]
    defer_count: int
    next_attempt_at: Optional[datetime]


def _normalize_timestamp(value: Optional[datetime]) -> Optional[datetime]:
    if value is None:
        return None
    if value.tzinfo is None:
        return value.replace(tzinfo=_LOCAL_TIMEZONE).astimezone(timezone.utc)
    return value.astimezone(timezone.utc)


def _encode_payload(payload: Any) -> str:
    if payload is None:
        payload = {}
    return json.dumps(payload, sort_keys=True, separators=(",", ":"))


def _decode_payload(payload: Any) -> Any:
    if payload is None:
        return {}
    if isinstance(payload, str):
        try:
            return json.loads(payload)
        except Exception:
            return payload
    return payload


def _derive_session_id(now: datetime) -> str:
    now_utc = _normalize_timestamp(now) or datetime.now(timezone.utc)
    return f"session-{now_utc.strftime('%Y%m%d')}"


def _next_session_run_index(
    con: duckdb.DuckDBPyConnection,
    *,
    pipeline_id: str,
    session_id: str,
) -> int:
    row = con.execute(
        """
        SELECT COALESCE(MAX(session_run_index), -1)
        FROM pipeline_runs
        WHERE pipeline_id = ? AND session_id = ?
        """,
        [pipeline_id, session_id],
    ).fetchone()
    if not row or row[0] is None:
        return 0
    return int(row[0]) + 1


class PipelineStateStore:
    """
    Pipeline runtime state store (non-RDF).

    This store manages the high-churn pipeline execution state in
    `cf-pipeline-states.duckdb`.
    """

    def __init__(self, db_path: Path, semantics_dir: Path, read_only: bool = False):
        self.db_path = Path(db_path)
        self.semantics_dir = Path(semantics_dir)
        self.read_only = bool(read_only)

    def _connect(self, *, read_only: Optional[bool] = None) -> duckdb.DuckDBPyConnection:
        ro = self.read_only if read_only is None else bool(read_only)
        return duckdb.connect(str(self.db_path), read_only=ro)

    def init_schema(self) -> None:
        """Create runtime-state tables and indexes when missing."""
        if not self.read_only:
            self.semantics_dir.mkdir(parents=True, exist_ok=True)

        def _table_columns(con: duckdb.DuckDBPyConnection, table_name: str) -> set[str]:
            rows = con.execute(f"PRAGMA table_info('{table_name}')").fetchall()
            cols: set[str] = set()
            for row in rows:
                # DuckDB PRAGMA table_info columns:
                # cid, name, type, notnull, dflt_value, pk
                if len(row) >= 2 and isinstance(row[1], str):
                    cols.add(row[1])
            return cols

        con = self._connect(read_only=False)
        try:
            con.execute(
                """
                CREATE TABLE IF NOT EXISTS pipeline_state (
                  pipeline_id TEXT PRIMARY KEY,
                  current_rev INTEGER NOT NULL,
                  desired_state TEXT NOT NULL,
                  concurrency_limit INTEGER DEFAULT 1,
                  priority INTEGER DEFAULT 0,
                  last_run_id TEXT,
                  last_run_at TIMESTAMP,
                  next_run_at TIMESTAMP,
                  created_at TIMESTAMP DEFAULT current_timestamp,
                  updated_at TIMESTAMP DEFAULT current_timestamp
                )
                """
            )
            con.execute(
                """
                CREATE TABLE IF NOT EXISTS pipeline_runs (
                  run_id TEXT PRIMARY KEY,
                  pipeline_id TEXT NOT NULL,
                  session_id TEXT NOT NULL,
                  session_run_index INTEGER NOT NULL,
                  pipeline_rev INTEGER NOT NULL,
                  status TEXT NOT NULL,
                  triggered_by_event_id TEXT,
                  created_at TIMESTAMP DEFAULT current_timestamp,
                  started_at TIMESTAMP,
                  finished_at TIMESTAMP,
                  termination_reason TEXT,
                  worker_id TEXT,
                  attempt INTEGER DEFAULT 0
                )
                """
            )
            con.execute(
                """
                CREATE TABLE IF NOT EXISTS run_events (
                  event_id TEXT PRIMARY KEY,
                  pipeline_id TEXT NOT NULL,
                  run_id TEXT,
                  event_type TEXT NOT NULL,
                  source TEXT NOT NULL DEFAULT 'system',
                  source_event_id TEXT,
                  payload JSON,
                  status TEXT NOT NULL DEFAULT 'pending',
                  created_at TIMESTAMP DEFAULT current_timestamp,
                  updated_at TIMESTAMP DEFAULT current_timestamp,
                  next_attempt_at TIMESTAMP,
                  defer_count INTEGER DEFAULT 0,
                  defer_reason TEXT,
                  last_error TEXT,
                  consumed_at TIMESTAMP,
                  dedupe_key TEXT
                )
                """
            )
            con.execute(
                """
                CREATE TABLE IF NOT EXISTS run_queue (
                  queue_id TEXT PRIMARY KEY,
                  pipeline_id TEXT NOT NULL,
                  run_id TEXT NOT NULL,
                  available_at TIMESTAMP NOT NULL,
                  priority INTEGER DEFAULT 0,
                  locked_by TEXT,
                  lock_expires_at TIMESTAMP,
                  attempts INTEGER DEFAULT 0,
                  last_error TEXT,
                  created_at TIMESTAMP DEFAULT current_timestamp,
                  updated_at TIMESTAMP DEFAULT current_timestamp
                )
                """
            )

            con.execute(
                """
                CREATE UNIQUE INDEX IF NOT EXISTS idx_run_events_dedupe
                ON run_events(dedupe_key)
                """
            )
            con.execute(
                """
                CREATE UNIQUE INDEX IF NOT EXISTS idx_run_queue_run
                ON run_queue(run_id)
                """
            )
            con.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_pipeline_runs_pipeline
                ON pipeline_runs(pipeline_id, status)
                """
            )
            con.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_run_events_pipeline
                ON run_events(pipeline_id)
                """
            )
            con.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_run_events_run
                ON run_events(run_id)
                """
            )
            con.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_run_events_pending
                ON run_events(pipeline_id, status, next_attempt_at, created_at)
                """
            )
            con.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_run_queue_available
                ON run_queue(available_at, priority)
                """
            )

            pipeline_runs_cols = _table_columns(con, "pipeline_runs")
            if "session_id" not in pipeline_runs_cols:
                con.execute("ALTER TABLE pipeline_runs ADD COLUMN session_id TEXT")
            if "session_run_index" not in pipeline_runs_cols:
                con.execute("ALTER TABLE pipeline_runs ADD COLUMN session_run_index INTEGER")

            run_events_cols = _table_columns(con, "run_events")
            if "source" not in run_events_cols:
                con.execute("ALTER TABLE run_events ADD COLUMN source TEXT DEFAULT 'system'")
            if "source_event_id" not in run_events_cols:
                con.execute("ALTER TABLE run_events ADD COLUMN source_event_id TEXT")
            if "status" not in run_events_cols:
                con.execute("ALTER TABLE run_events ADD COLUMN status TEXT DEFAULT 'pending'")
            if "updated_at" not in run_events_cols:
                con.execute("ALTER TABLE run_events ADD COLUMN updated_at TIMESTAMP DEFAULT current_timestamp")
            if "next_attempt_at" not in run_events_cols:
                con.execute("ALTER TABLE run_events ADD COLUMN next_attempt_at TIMESTAMP")
            if "defer_count" not in run_events_cols:
                con.execute("ALTER TABLE run_events ADD COLUMN defer_count INTEGER DEFAULT 0")
            if "defer_reason" not in run_events_cols:
                con.execute("ALTER TABLE run_events ADD COLUMN defer_reason TEXT")
            if "last_error" not in run_events_cols:
                con.execute("ALTER TABLE run_events ADD COLUMN last_error TEXT")
            if "consumed_at" not in run_events_cols:
                con.execute("ALTER TABLE run_events ADD COLUMN consumed_at TIMESTAMP")

            con.execute(
                """
                UPDATE pipeline_runs
                SET session_id = COALESCE(session_id, pipeline_id || ':legacy-session')
                WHERE session_id IS NULL OR TRIM(session_id) = ''
                """
            )
            con.execute(
                """
                UPDATE pipeline_runs AS pr
                SET session_run_index = ranked.session_run_index
                FROM (
                  SELECT
                    run_id,
                    ROW_NUMBER() OVER (
                      PARTITION BY pipeline_id, session_id
                      ORDER BY created_at, run_id
                    ) - 1 AS session_run_index
                  FROM pipeline_runs
                ) AS ranked
                WHERE pr.run_id = ranked.run_id
                  AND pr.session_run_index IS NULL
                """
            )
            con.execute(
                """
                UPDATE run_events
                SET source = COALESCE(NULLIF(TRIM(source), ''), 'system')
                WHERE source IS NULL OR TRIM(source) = ''
                """
            )
            con.execute(
                """
                UPDATE run_events
                SET status = CASE
                  WHEN run_id IS NULL THEN 'pending'
                  ELSE 'consumed'
                END
                WHERE status IS NULL OR TRIM(status) = ''
                """
            )
            con.execute(
                """
                UPDATE run_events
                SET updated_at = COALESCE(updated_at, created_at, current_timestamp)
                WHERE updated_at IS NULL
                """
            )
            con.execute(
                """
                UPDATE run_events
                SET consumed_at = COALESCE(consumed_at, updated_at, created_at, current_timestamp)
                WHERE run_id IS NOT NULL AND consumed_at IS NULL
                """
            )
            con.execute(
                """
                CREATE INDEX IF NOT EXISTS idx_pipeline_runs_session
                ON pipeline_runs(pipeline_id, session_id, session_run_index)
                """
            )
        finally:
            con.close()

    def insert_event(
        self,
        *,
        pipeline_id: str,
        event_type: str,
        payload: Any = None,
        source: str = "external",
        dedupe_key: Optional[str] = None,
        source_event_id: Optional[str] = None,
        con: Optional[duckdb.DuckDBPyConnection] = None,
    ) -> tuple[str, bool]:
        own_connection = con is None
        db = con or self._connect(read_only=False)
        norm_source = (source or "external").strip() or "external"
        norm_dedupe = dedupe_key.strip() if isinstance(dedupe_key, str) and dedupe_key.strip() else None
        event_id = str(uuid.uuid4())
        payload_json = _encode_payload(payload)
        try:
            if norm_dedupe:
                existing = db.execute(
                    "SELECT event_id FROM run_events WHERE dedupe_key = ? LIMIT 1",
                    [norm_dedupe],
                ).fetchone()
                if existing and existing[0]:
                    return str(existing[0]), False

            try:
                db.execute(
                    """
                    INSERT INTO run_events (
                      event_id,
                      pipeline_id,
                      run_id,
                      event_type,
                      source,
                      source_event_id,
                      payload,
                      status,
                      created_at,
                      updated_at,
                      dedupe_key
                    )
                    VALUES (
                      ?,
                      ?,
                      NULL,
                      ?,
                      ?,
                      ?,
                      CAST(? AS JSON),
                      'pending',
                      current_timestamp,
                      current_timestamp,
                      ?
                    )
                    """,
                    [
                        event_id,
                        pipeline_id,
                        event_type,
                        norm_source,
                        source_event_id,
                        payload_json,
                        norm_dedupe,
                    ],
                )
            except duckdb.ConstraintException:
                if norm_dedupe:
                    existing = db.execute(
                        "SELECT event_id FROM run_events WHERE dedupe_key = ? LIMIT 1",
                        [norm_dedupe],
                    ).fetchone()
                    if existing and existing[0]:
                        return str(existing[0]), False
                raise
            return event_id, True
        finally:
            if own_connection:
                db.close()

    def fetch_next_pending_event(
        self,
        *,
        pipeline_id: Optional[str] = None,
        now: Optional[datetime] = None,
        con: Optional[duckdb.DuckDBPyConnection] = None,
    ) -> Optional[PendingRunEvent]:
        own_connection = con is None
        db = con or self._connect(read_only=False)
        now_utc = _normalize_timestamp(now) or datetime.now(timezone.utc)
        try:
            query = """
                SELECT
                  event_id,
                  pipeline_id,
                  event_type,
                  source,
                  source_event_id,
                  payload,
                  created_at,
                  dedupe_key,
                  defer_count,
                  next_attempt_at
                FROM run_events
                WHERE run_id IS NULL
                  AND status = 'pending'
                  AND (next_attempt_at IS NULL OR next_attempt_at <= ?)
            """
            params: list[Any] = [now_utc]
            if pipeline_id:
                query += " AND pipeline_id = ?"
                params.append(pipeline_id)
            query += """
                ORDER BY COALESCE(next_attempt_at, created_at), created_at, event_id
                LIMIT 1
            """
            row = db.execute(query, params).fetchone()
            if not row:
                return None
            return PendingRunEvent(
                event_id=str(row[0]),
                pipeline_id=str(row[1]),
                event_type=str(row[2]),
                source=str(row[3] or "system"),
                source_event_id=str(row[4]) if row[4] is not None else None,
                payload=_decode_payload(row[5]),
                created_at=_normalize_timestamp(row[6]),
                dedupe_key=str(row[7]) if row[7] is not None else None,
                defer_count=int(row[8] or 0),
                next_attempt_at=_normalize_timestamp(row[9]),
            )
        finally:
            if own_connection:
                db.close()

    def defer_event(
        self,
        *,
        event_id: str,
        reason: str,
        next_attempt_at: datetime,
        last_error: Optional[str] = None,
        con: Optional[duckdb.DuckDBPyConnection] = None,
    ) -> bool:
        own_connection = con is None
        db = con or self._connect(read_only=False)
        next_attempt = _normalize_timestamp(next_attempt_at) or datetime.now(timezone.utc)
        try:
            row = db.execute(
                """
                SELECT
                  pipeline_id,
                  event_type,
                  source,
                  source_event_id,
                  payload,
                  created_at,
                  dedupe_key,
                  COALESCE(defer_count, 0)
                FROM run_events
                WHERE event_id = ?
                  AND run_id IS NULL
                LIMIT 1
                """,
                [event_id],
            ).fetchone()
            if not row:
                return False
            db.execute(
                "DELETE FROM run_events WHERE event_id = ? AND run_id IS NULL",
                [event_id],
            )
            db.execute(
                """
                INSERT INTO run_events (
                  event_id,
                  pipeline_id,
                  run_id,
                  event_type,
                  source,
                  source_event_id,
                  payload,
                  status,
                  created_at,
                  updated_at,
                  next_attempt_at,
                  defer_count,
                  defer_reason,
                  last_error,
                  consumed_at,
                  dedupe_key
                )
                VALUES (
                  ?,
                  ?,
                  NULL,
                  ?,
                  ?,
                  ?,
                  CAST(? AS JSON),
                  'pending',
                  ?,
                  current_timestamp,
                  ?,
                  ?,
                  ?,
                  ?,
                  NULL,
                  ?
                )
                """,
                [
                    event_id,
                    str(row[0]),
                    str(row[1]),
                    str(row[2] or "system"),
                    str(row[3]) if row[3] is not None else None,
                    _encode_payload(_decode_payload(row[4])),
                    row[5],
                    next_attempt,
                    int(row[7] or 0) + 1,
                    reason,
                    last_error,
                    str(row[6]) if row[6] is not None else None,
                ],
            )
            return True
        finally:
            if own_connection:
                db.close()

    def consume_event_to_run(
        self,
        *,
        event_id: str,
        pipeline_rev: int,
        priority: int,
        now: Optional[datetime] = None,
        con: Optional[duckdb.DuckDBPyConnection] = None,
    ) -> Optional[tuple[str, str, str]]:
        own_connection = con is None
        db = con or self._connect(read_only=False)
        now_utc = _normalize_timestamp(now) or datetime.now(timezone.utc)
        run_id = str(uuid.uuid4())
        queue_id = str(uuid.uuid4())
        try:
            candidate = db.execute(
                """
                SELECT
                  pipeline_id,
                  event_type,
                  source,
                  source_event_id,
                  payload,
                  created_at,
                  dedupe_key
                FROM run_events
                WHERE event_id = ?
                  AND run_id IS NULL
                  AND status = 'pending'
                LIMIT 1
                """,
                [event_id],
            ).fetchone()
            if not candidate or candidate[0] is None:
                return None
            pipeline_id = str(candidate[0])

            db.execute(
                "DELETE FROM run_events WHERE event_id = ? AND run_id IS NULL AND status = 'pending'",
                [event_id],
            )
            still_pending = db.execute(
                """
                SELECT 1
                FROM run_events
                WHERE event_id = ?
                  AND run_id IS NULL
                  AND status = 'pending'
                LIMIT 1
                """,
                [event_id],
            ).fetchone()
            if still_pending:
                return None

            db.execute(
                """
                INSERT INTO run_events (
                  event_id,
                  pipeline_id,
                  run_id,
                  event_type,
                  source,
                  source_event_id,
                  payload,
                  status,
                  created_at,
                  updated_at,
                  next_attempt_at,
                  defer_count,
                  defer_reason,
                  last_error,
                  consumed_at,
                  dedupe_key
                )
                VALUES (
                  ?,
                  ?,
                  ?,
                  ?,
                  ?,
                  ?,
                  CAST(? AS JSON),
                  'consumed',
                  ?,
                  current_timestamp,
                  NULL,
                  0,
                  NULL,
                  NULL,
                  current_timestamp,
                  ?
                )
                """,
                [
                    event_id,
                    pipeline_id,
                    run_id,
                    str(candidate[1]),
                    str(candidate[2] or "system"),
                    str(candidate[3]) if candidate[3] is not None else None,
                    _encode_payload(_decode_payload(candidate[4])),
                    candidate[5],
                    str(candidate[6]) if candidate[6] is not None else None,
                ],
            )
            consumed_check = db.execute(
                """
                SELECT pipeline_id
                FROM run_events
                WHERE event_id = ? AND run_id = ?
                LIMIT 1
                """,
                [event_id, run_id],
            ).fetchone()
            if not consumed_check:
                return None

            session_id = _derive_session_id(now_utc)
            session_run_index = _next_session_run_index(
                db,
                pipeline_id=pipeline_id,
                session_id=session_id,
            )

            db.execute(
                """
                INSERT INTO pipeline_runs (
                  run_id,
                  pipeline_id,
                  session_id,
                  session_run_index,
                  pipeline_rev,
                  status,
                  triggered_by_event_id,
                  created_at
                )
                VALUES (?, ?, ?, ?, ?, ?, ?, current_timestamp)
                """,
                [
                    run_id,
                    pipeline_id,
                    session_id,
                    int(session_run_index),
                    int(pipeline_rev),
                    "queued",
                    event_id,
                ],
            )
            db.execute(
                """
                INSERT INTO run_queue (
                  queue_id,
                  pipeline_id,
                  run_id,
                  available_at,
                  priority,
                  created_at,
                  updated_at
                )
                VALUES (?, ?, ?, ?, ?, current_timestamp, current_timestamp)
                """,
                [queue_id, pipeline_id, run_id, now_utc, int(priority)],
            )
            registration = db.execute(
                """
                SELECT
                  re.run_id,
                  pr.triggered_by_event_id,
                  rq.queue_id
                FROM run_events AS re
                JOIN pipeline_runs AS pr
                  ON pr.run_id = re.run_id
                JOIN run_queue AS rq
                  ON rq.run_id = pr.run_id
                WHERE re.event_id = ?
                  AND re.run_id = ?
                LIMIT 1
                """,
                [event_id, run_id],
            ).fetchone()
            if not registration:
                return None
            if str(registration[1] or "") != event_id:
                return None
            if str(registration[2] or "") != queue_id:
                return None
            return run_id, queue_id, pipeline_id
        finally:
            if own_connection:
                db.close()



from __future__ import annotations

import json
from pathlib import Path

import duckdb
from rdflib import Dataset

from cf_ontology import ingest_rdf_documents
from cf_ontology.rdf_document import load_rdf_document
from cf_ontology.semantics_paths import resolve_semantics_db_path
from cf_ontology.signature import generate_signature_header

_STRUCTURED_FORMAT = "json" + "-ld"


def _repo_root() -> Path:
    return Path(__file__).resolve().parents[3]


def _write_nquads(path: Path, document: dict) -> Path:
    dataset = Dataset()
    dataset.parse(data=json.dumps(document), format=_STRUCTURED_FORMAT)
    payload = dataset.serialize(format="nquads")
    if isinstance(payload, bytes):
        payload = payload.decode("utf-8")
    path.write_text(payload, encoding="utf-8")
    return path


def _basic_signal_steps_document() -> dict:
    source = (
        _repo_root()
        / "sandcastle"
        / "cf_basic_steps"
        / "cf_basic_signal"
        / "src"
        / "cf_basic_signal"
        / "steps.nq"
    )
    return load_rdf_document(source, label="step document")


def test_nquads_step_document_ingest_stores_expected_graph_ids(tmp_path: Path) -> None:
    document = _basic_signal_steps_document()
    nq_path = _write_nquads(tmp_path / "basic_signal_steps.nq", document)
    semantics_dir = tmp_path / "semantics"

    ingest_rdf_documents(
        [nq_path],
        semantics_dir=semantics_dir,
        graph_type="step",
        package="cf.basic.signal",
    )

    db_path = resolve_semantics_db_path(semantics_dir, kind="steps")
    con = duckdb.connect(str(db_path), read_only=True)
    try:
        rows = con.execute("SELECT graph_id FROM graphs ORDER BY graph_id").fetchall()
    finally:
        con.close()

    graph_ids = {str(row[0]) for row in rows}
    assert any(graph_id.endswith("basic-signal/AverageStep") for graph_id in graph_ids)
    assert any(graph_id.endswith("basic-signal/FifoWindowBufferStep") for graph_id in graph_ids)


def test_siggen_accepts_nquads_step_document(tmp_path: Path) -> None:
    document = _basic_signal_steps_document()
    nq_path = _write_nquads(tmp_path / "basic_signal_steps.nq", document)
    out_path = tmp_path / "signatures.hpp"

    signatures = generate_signature_header(
        documents=[nq_path],
        out_path=out_path,
        scratch=True,
    )

    assert signatures
    assert any(signature.step_iri.endswith("AverageStep") for signature in signatures)
    header = out_path.read_text(encoding="utf-8")
    assert "AverageStep" in header